==> vtorch-1.0.1-py3-none-any.whl: vtorch/distillation/utils.py <==
import math
from functools import partial
from typing import Optional, Tuple

import torch
from torch import nn
from transformers import PreTrainedModel
from transformers.models.bert.modeling_bert import BertSelfAttention


def forward(
    self: BertSelfAttention,
    hidden_states: torch.Tensor,
    attention_mask: Optional[torch.Tensor] = None,
    head_mask: Optional[torch.Tensor] = None,
    encoder_hidden_states: Optional[torch.Tensor] = None,
    encoder_attention_mask: Optional[torch.Tensor] = None,
    past_key_value: Optional[Tuple[torch.Tensor, torch.Tensor]] = None,
    output_attentions: bool = False,
) -> Tuple[torch.Tensor, ...]:
    mixed_query_layer = self.query(hidden_states)

    # If this is instantiated as a cross-attention module, the keys
    # and values come from an encoder; the attention mask needs to be
    # such that the encoder's padding tokens are not attended to.
    is_cross_attention = encoder_hidden_states is not None

    if is_cross_attention and past_key_value is not None:
        # reuse k,v, cross_attentions
        key_layer = past_key_value[0]
        value_layer = past_key_value[1]
        attention_mask = encoder_attention_mask
    elif is_cross_attention:
        key_layer = self.transpose_for_scores(self.key(encoder_hidden_states))
        value_layer = self.transpose_for_scores(self.value(encoder_hidden_states))
        attention_mask = encoder_attention_mask
    elif past_key_value is not None:
        key_layer = self.transpose_for_scores(self.key(hidden_states))
        value_layer = self.transpose_for_scores(self.value(hidden_states))
        key_layer = torch.cat([past_key_value[0], key_layer], dim=2)
        value_layer = torch.cat([past_key_value[1], value_layer], dim=2)
    else:
        key_layer = self.transpose_for_scores(self.key(hidden_states))
        value_layer = self.transpose_for_scores(self.value(hidden_states))

    query_layer = self.transpose_for_scores(mixed_query_layer)

    if self.is_decoder:
        # if cross_attention save Tuple(torch.Tensor, torch.Tensor) of all cross attention key/value_states.
        # Further calls to cross_attention layer can then reuse all cross-attention
        # key/value_states (first "if" case)
        # if uni-directional self-attention (decoder) save Tuple(torch.Tensor, torch.Tensor) of
        # all previous decoder key/value_states. Further calls to uni-directional self-attention
        # can concat previous decoder key/value_states to current projected key/value_states (third "elif" case)
        # if encoder bi-directional self-attention `past_key_value` is always `None`
        past_key_value = (key_layer, value_layer)

    # Take the dot product between "query" and "key" to get the raw attention scores.
    attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))

    if self.position_embedding_type == "relative_key" or self.position_embedding_type == "relative_key_query":
        seq_length = hidden_states.size()[1]
        position_ids_l = torch.arange(seq_length, dtype=torch.long, device=hidden_states.device).view(-1, 1)
        position_ids_r = torch.arange(seq_length, dtype=torch.long, device=hidden_states.device).view(1, -1)
        distance = position_ids_l - position_ids_r
        positional_embedding = self.distance_embedding(distance + self.max_position_embeddings - 1)
        positional_embedding = positional_embedding.to(dtype=query_layer.dtype)  # fp16 compatibility

        if self.position_embedding_type == "relative_key":
            relative_position_scores = torch.einsum("bhld,lrd->bhlr", query_layer, positional_embedding)
            attention_scores = attention_scores + relative_position_scores
        elif self.position_embedding_type == "relative_key_query":
            relative_position_scores_query = torch.einsum("bhld,lrd->bhlr", query_layer, positional_embedding)
            relative_position_scores_key = torch.einsum("bhrd,lrd->bhlr", key_layer, positional_embedding)
            attention_scores = attention_scores + relative_position_scores_query + relative_position_scores_key

    attention_scores = attention_scores / math.sqrt(self.attention_head_size)
    if attention_mask is not None:
        # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
        attention_scores = attention_scores + attention_mask

    # Normalize the attention scores to probabilities.
    attention_probs = nn.Softmax(dim=-1)(attention_scores)

    # This is actually dropping out entire tokens to attend to, which might
    # seem a bit unusual, but is taken from the original Transformer paper.
    attention_probs = self.dropout(attention_probs)

    # Mask heads if we want to
    if head_mask is not None:
        attention_probs = attention_probs * head_mask

    context_layer = torch.matmul(attention_probs, value_layer)

    context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
    new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
    context_layer = context_layer.view(*new_context_layer_shape)

    outputs = (context_layer, attention_scores) if output_attentions else (context_layer,)

    if self.is_decoder:
        outputs = outputs + (past_key_value,)  # type: ignore
    return outputs  # type: ignore


def patch_self_attention_forward(model: PreTrainedModel) -> None:
    """
    Function to patch model's self_attention forward method to receive the `attention_scores` instead of
    `attention_probs` in the output. There is assumption that we will use BertModel architecture.
    """
    if hasattr(model, "encoder"):
        encoder = model.encoder
    else:
        raise AttributeError("The encoder can't be found in the model by 'encoder' attribute")

    if hasattr(encoder, "layer") and isinstance(encoder.layer, nn.ModuleList):
        layers = encoder.layer
    else:
        raise AttributeError("Transformer layers can't be found in the encoder by 'layer' attribute")

    if not hasattr(next(iter(layers)), "attention"):
        raise AttributeError("Attention block can't be found in transformer layers by 'attention' attribute")
    elif not hasattr(next(iter(layers)).attention, "self"):
        raise AttributeError("MultiHeadSelfAttention block can't be found in attention block by 'self' attribute")
    else:
        for layer in layers:
            layer.attention.self.forward = partial(forward, layer.attention.self)

==> vtorch-1.4.1-py3-none-any.whl: vtorch/distillation/utils.py <==
import math
from functools import partial
from typing import Optional, Tuple

import torch
from torch import nn
from transformers import PreTrainedModel
from transformers.modeling_bert import BertSelfAttention


def forward(
    self: BertSelfAttention,
    hidden_states: torch.Tensor,
    attention_mask: Optional[torch.Tensor] = None,
    head_mask: Optional[torch.Tensor] = None,
    encoder_hidden_states: Optional[torch.Tensor] = None,
    encoder_attention_mask: Optional[torch.Tensor] = None,
) -> Tuple[torch.Tensor, ...]:
    mixed_query_layer = self.query(hidden_states)

    # If this is instantiated as a cross-attention module, the keys
    # and values come from an encoder; the attention mask needs to be
    # such that the encoder's padding tokens are not attended to.
    if encoder_hidden_states is not None:
        mixed_key_layer = self.key(encoder_hidden_states)
        mixed_value_layer = self.value(encoder_hidden_states)
        attention_mask = encoder_attention_mask
    else:
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

    query_layer = self.transpose_for_scores(mixed_query_layer)
    key_layer = self.transpose_for_scores(mixed_key_layer)
    value_layer = self.transpose_for_scores(mixed_value_layer)

    # Take the dot product between "query" and "key" to get the raw attention scores.
    attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
    attention_scores = attention_scores / math.sqrt(self.attention_head_size)
    if attention_mask is not None:
        # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
        attention_scores = attention_scores + attention_mask

    # Normalize the attention scores to probabilities.
    attention_probs = nn.Softmax(dim=-1)(attention_scores)

    # This is actually dropping out entire tokens to attend to, which might
    # seem a bit unusual, but is taken from the original Transformer paper.
    attention_probs = self.dropout(attention_probs)

    # Mask heads if we want to
    if head_mask is not None:
        attention_probs = attention_probs * head_mask

    context_layer = torch.matmul(attention_probs, value_layer)

    context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
    new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
    context_layer = context_layer.view(*new_context_layer_shape)

    outputs = (context_layer, attention_scores) if self.output_attentions else (context_layer,)
    return outputs  # type: ignore


def patch_self_attention_forward(model: PreTrainedModel) -> None:
    """
    Function to patch model's self_attention forward method to receive the `attention_scores` instead of
    `attention_probs` in the output. There is assumption that we will use BertModel architecture.
    """
    if hasattr(model, "encoder"):
        encoder = model.encoder
    else:
        raise AttributeError("The encoder can't be found in the model by 'encoder' attribute")

    if hasattr(encoder, "layer") and isinstance(encoder.layer, nn.ModuleList):
        layers = encoder.layer
    else:
        raise AttributeError("Transformer layers can't be found in the encoder by 'layer' attribute")

    if not hasattr(next(iter(layers)), "attention"):
        raise AttributeError("Attention block can't be found in transformer layers by 'attention' attribute")
    elif not hasattr(next(iter(layers)).attention, "self"):
        raise AttributeError("MultiHeadSelfAttention block can't be found in attention block by 'self' attribute")
    else:
        for layer in layers:
            layer.attention.self.forward = partial(forward, layer.attention.self)
